/* Pavel Surynek, 2008 - 2009 */
/* pavel.surynek@mff.cuni.cz */
/* mit_main.cpp */

/*
  An implementation of the algorithm for solving sliding box puzzles by
  Kornhauser, Miler, and Spirakis.
 */

#include "slider.h"
#include "mit.h"
#include "tbox.h"
#include "obox.h"


SLIDER slider1, slider2, slider3;
SOLUTION_DATABASE database;


int main(int UNUSED(argc), char **UNUSED(argv))
{
	clock_t begin, end;

//	generate_solution_database_Revised("solution_database_revised_trans");
//	generate_solution_3_database_Revised("solution_database_revised_3_cycle");

//	init_slider(0, 7, 5, 10, &slider1);

	FILE *f;
	f = fopen("Results_3_Theta_vs_MIT_xaos.txt", "w");

	database.load("solution_database_T.txt");

//      for (int l = 0; l < 100; ++l)

	for (int a = 3; a < 8; ++a)
	{
		for (int b = 2; b < 4; ++b)
		{
			for (int c = 3; c < 8; ++c)
			{
				if ((a + b) % 2 == 0 && (b + c) % 2 == 0 && (a + c) % 2 == 0)
				{
					continue;
				}
 
				printf("ABC:%d,%d,%d\n", a, b, c);
				fprintf(f, "%d\t%d\t%d\t", a, b, c);

				FINE_SOLUTION solution, mit_solution;
				
				begin = clock();

				init_slider(a, b, c, &slider1);
				
				for (int i = 0; i < slider1.m_V * slider1.m_V; ++i)
				{
					int pos1 = rand() % (slider1.m_V - 1) + 1;
					int pos2 = (pos1 + 1) % (slider1.m_V - 1) + 1;
					
					swap_goal(pos1, pos2, &slider1);
				}
				
				bool in_database;
				
				copy_slider(&slider1, &slider2);
				copy_slider(&slider1, &slider3);

//		solve_slider_tbox(&slider1, database, solution, in_database, 1000.0);
//		solve_slider_tbox_xaos(&slider1, database, solution, in_database, 1000.0);
//				solve_slider_3_theta(&slider1, database, solution, 1000.0);
		solve_slider_3_tbox_xaos(&slider1, database, solution, in_database, 1000.0);

				solve_slider_mit(&slider3, mit_solution);
	
				printf("TBOX solution length: %d\n", solution.size());
				printf(" MIT solution length: %d\n", mit_solution.size());
				printf("         Improvement: %.3f\n", (double)mit_solution.size() / (double)solution.size());
/*
				print_slider(&slider2);
				execute_fine_solution(&slider2, solution);
				print_slider(&slider2);
*/
				destroy_slider(&slider1);
				destroy_slider(&slider2);
				destroy_slider(&slider3);
				
				end = clock();
				printf("Time:%.3f\n", (end - begin) / (double)CLOCKS_PER_SEC);

				fprintf(f, "%d\t%d\t%.3f\t%.3f\n", solution.size(), mit_solution.size(), (double)mit_solution.size() / (double)solution.size(), (end - begin) / (double)CLOCKS_PER_SEC);
				fflush(f);

				database.save("solution_database_T.txt");
			}
		}
	}

//	database.save("solution_database_A.txt");

	return 0;
}
